##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module contains the generic parser definition.
"""

# Imports
import re
import warnings
from string import Formatter
from itertools import product
from caravel.parsers import BIDSParser


# Global parameters
# > define all the available parsers
PARSERS = [BIDSParser]


def get_parser(project, confdir, layoutdir):
    """ Method to return the appropriate parser for your study.

    Parameters
    ----------
    project: str
        the name of the project you are working on.
    confdir: str
        the locations of the configuration file of the current project.
    layoutdir: str
        the location of the pre-generated parsing representations. If None
        switch to managers mode.

    Returns
    -------
    parser: obj
        the dataset appropriate parser.
    """
    for parser_class in PARSERS:
        parser = parser_class(project, confdir, layoutdir)
        if parser.can_load():
            return parser
    raise ValueError(f"No loader available for '{project}'.")


def listify(obj):
    """ Wraps all non-list or tuple objects in a list.
    """
    return obj if isinstance(obj, (list, tuple, type(None))) else [obj]


def _expand_keys(entities):
    """ Generate multiple replacement queries based on all combinations
    of values.

    Examples
    --------
    >>> entities = {'subject': ['01', '02'], 'session': ['1', '2'],
                    'task': ['rest', 'finger']}
    >>> out = _expand_entities(entities)
    >>> len(out)
    8
    >>> {'subject': '01', 'session': '1', 'task': 'rest'} in out
    True
    >>> {'subject': '02', 'session': '1', 'task': 'rest'} in out
    True
    >>> {'subject': '01', 'session': '2', 'task': 'rest'} in out
    True
    >>> {'subject': '02', 'session': '2', 'task': 'rest'} in out
    True
    >>> {'subject': '01', 'session': '1', 'task': 'finger'} in out
    True
    >>> {'subject': '02', 'session': '1', 'task': 'finger'} in out
    True
    >>> {'subject': '01', 'session': '2', 'task': 'finger'} in out
    True
    >>> {'subject': '02', 'session': '2', 'task': 'finger'} in out
    True
    """
    keys = list(entities.keys())
    values = list(product(*[entities[k] for k in keys]))
    return [dict(zip(keys, combs)) for combs in values]


_PATTERN_FIND = re.compile(
    r"({([\w\d]*?)(?:<([^>]+)>)?(?:\|((?:\.?[\w])+))?\})")


def build_path(keys, path_patterns, strict=False):
    """ Constructs a path given a set of keys and a list of potential filename
    patterns to use.

    Parameters
    ----------
    keys: dict
        A dictionary mapping key names to key values.
        None key or empty-string key will be removed.
        Otherwise, kets will be cast to string values, therefore
        if any format is expected (e.g., zero-padded integers), the
        value should be formatted.
    path_patterns: str or list of str
        One or more filename patterns to write the file to.
        Keys should be represented by the name surrounded by curly braces.
        Optional portions of the patterns should be denoted by square brackets.
        Keys that require a specific value for the pattern to match can pass
        them inside angle brackets.
    strict: bool
        If True, all passed keys must be matched inside a pattern in order to
        be a valid match.
        If False, extra keys will be ignored so long as all mandatory keys are
        found.

    Returns
    -------
    path: str
        A constructed path for this file based on the provided patterns, or
        None if no path was built given the combination of keys and patterns.
    """
    path_patterns = listify(path_patterns)
    keys = {key: listify(val) for key, val in keys.items() if val or val == 0}

    # Loop over available patterns, return first one that matches all
    for pattern in path_patterns:
        keys_matched = list(_PATTERN_FIND.findall(pattern))
        defined = [elem[1] for elem in keys_matched]

        # If strict, all keys must be contained in the pattern
        if strict and set(keys.keys()) - set(defined):
            continue

        # Iterate through the provided path patterns
        new_path = pattern
        # Expand options within valid values and
        # check whether keys provided have acceptable value
        tmp_keys = keys.copy()
        for fmt, name, valid, defval in keys_matched:
            valid_expanded = valid.split('|')
            if valid_expanded and defval and defval not in valid_expanded:
                warnings.warn(f"Pattern '{fmt}' is inconsistent as it defines "
                              "an invalid default value.")
            if (valid_expanded and name in keys and
                    set(keys[name]) - set(valid_expanded)):
                continue
            if defval and name not in tmp_keys:
                tmp_keys[name] = [defval]

            # At this point, valid & default values are checked &
            # set - simplify pattern
            new_path = new_path.replace(fmt, f'{{{name}}}')

        optional_patterns = re.findall(r'(\[.*?\])', new_path)
        # Optional patterns with selector are cast to mandatory or removed
        for op in optional_patterns:
            for ent_name in {k for k, v in keys.items() if (v is not None)
                             and not (v[0] != v[0])}:
                if (f'{{{ent_name}}}') in op:
                    new_path = new_path.replace(op, op[1:-1])
                    continue

            # Surviving optional patterns are removed
            new_path = new_path.replace(op, '')
        # Replace keys
        fields = {pat[1] for pat in Formatter().parse(new_path)
                  if pat[1] and not pat[1].isdigit()}
        if fields - set(tmp_keys.keys()):
            continue

        tmp_keys = {k: v for k, v in tmp_keys.items()
                    if k in fields}
        new_path = [
            new_path.format(**e)
            for e in _expand_keys(tmp_keys)
        ]
        if new_path:
            if len(new_path) == 1:
                new_path = new_path[0]
            return new_path

    return None



##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module defines the TARGZ dataset tester.
"""

# Third party import
import gzip

# Package import
from .loader_base import LoaderBase


class TARGZ(LoaderBase):
    """ Define the TARGZ tester.
    """
    allowed_extensions = [".tar.gz", ]

    def load(self, path):
        """ A method that test a TARGZ file.

        Parameters
        ----------
        path: str
            the path to the TARGZ file.

        Returns
        -------
        data: object
            the loaded data.
        """
        with gzip.open(path, "rb") as open_file:
            for line in open_file:
                pass



from ..base import WithRequester


class FederatedCloudShare(WithRequester):
    API_URL = "/ocs/v2.php/apps/files_sharing/api/v1"
    FEDERATED = "remote_shares"

    def get_federated_url(self, additional_url=""):
        if additional_url:
            return f"{self.FEDERATED}/{additional_url}"
        return self.FEDERATED

    def list_accepted_federated_cloudshares(self):
        url = self.get_federated_url()
        return self.requester.get(url)

    def get_known_federated_cloudshare(self, sid):
        url = self.get_federated_url(sid)
        return self.requester.get(url)

    def delete_accepted_federated_cloudshare(self, sid):
        url = self.get_federated_url(sid)
        return self.requester.delete(url)

    def list_pending_federated_cloudshares(self, sid):
        url = self.get_federated_url("pending")
        return self.requester.get(url)

    def accept_pending_federated_cloudshare(self, sid):
        url = self.get_federated_url(f"pending/{sid}")
        return self.requester.post(url)

    def decline_pending_federated_cloudshare(self, sid):
        url = self.get_federated_url(f"pending/{sid}")
        return self.requester.delete(url)



from ..base import WithRequester


class Group(WithRequester):
    API_URL = "/ocs/v1.php/cloud/groups"
    SUCCESS_CODE = 100

    def get_groups(self, search=None, limit=None, offset=None):
        """ Retrieve a list of groups from the Nextcloud server.
        """
        params = {
            'search': search,
            'limit': limit,
            'offset': offset
        }
        return self.requester.get(params=params)

    def add_group(self, gid):
        """ Add a new group.
        """
        msg = {"groupid": gid}
        return self.requester.post("", msg)

    def get_group(self, gid):
        """ Retrieve a list of group members.
        """
        return self.requester.get(f"{gid}")

    def get_subadmins(self, gid):
        """ List subadmins of the group.
        """
        return self.requester.get(f"{gid}/subadmins")

    def delete_group(self, gid):
        """ Remove a group.
        """
        return self.requester.delete(f"{gid}")



from ..base import WithRequester


class Notifications(WithRequester):
    API_URL = "/ocs/v2.php/apps/notifications/api/v2/notifications"
    SUCCESS_CODE = 200

    def get_notifications(self):
        """ Get list of notifications for a logged in user.
        """
        return self.requester.get()

    def get_notification(self, notification_id):
        """ Get single notification by id for a user.
        """
        return self.requester.get(url=notification_id)

    def delete_notification(self, notification_id):
        """ Delete single notification by id for a user.
        """
        return self.requester.delete(url=notification_id)

    def delete_all_notifications(self):
        """ Delete all notification for a logged in user.
        """
        return self.requester.delete()



##########################################################################
# NSAp - Copyright (C) CEA, 2020
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module defines the png dataset loader.
"""

# Third party import
import imageio

# Package import
from .loader_base import LoaderBase


class PNG(LoaderBase):
    """ Define the PNG loader.
    """
    allowed_extensions = [".png"]

    def load(self, path):
        """ A method that load the png data.

        Parameters
        ----------
        path: str
            the path to the png file to be loaded.

        Returns
        -------
        data: imageio numpy array
            the loaded image.
        """
        return imageio.imread(path)

    def save(self, data, outpath):
        """ A method that save the image in png.

        Parameters
        ----------
        data: imageio numpy array
            the image to be saved.
        outpath: str
            the path where the the png image will be saved.
        """
        imageio.imwrite(outpath, data)



##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module defines common loaders for imaging-genetic datasets.
"""

from ._nifti import NIFTI
from ._tsv import TSV
from ._json import JSON
from ._dwi import DWI
from ._targz import TARGZ
from ._png import PNG
from ._csv import CSV
from ._xlsx import XLSX
from ._pdf import PDF
from ._plink import PLINK
from ._vcf import VCF
from ._mp4 import MP4
from ._edf import EDF
from ._mzml import MZML



import re
import os
import xml.etree.ElementTree as ET
from ..base import WithRequester


class WebDAV(WithRequester):
    API_URL = "/remote.php/dav/files"

    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.json_output = kwargs.get('json_output')

    def list_folders(self, uid, path=None, depth=1, all_properties=False):
        """ Get path files list with files properties for given user, with
        given depth.
        """
        if all_properties:
            data = """<?xml version="1.0"?>
                <d:propfind xmlns:d="DAV:" xmlns:oc="http://owncloud.org/ns"
                            xmlns:nc="http://nextcloud.org/ns">
                  <d:prop>
                        <d:getlastmodified />
                        <d:getetag />
                        <d:getcontenttype />
                        <d:resourcetype />
                        <oc:fileid />
                        <oc:permissions />
                        <oc:size />
                        <d:getcontentlength />
                        <nc:has-preview />
                        <oc:favorite />
                        <oc:comments-unread />
                        <oc:owner-display-name />
                        <oc:share-types />
                  </d:prop>
                </d:propfind>
            """
        else:
            data = None
        additional_url = uid
        if path:
            additional_url = f"{additional_url}/{path}"
        resp = self.requester.propfind(additional_url=additional_url,
                                       headers={"Depth": str(depth)},
                                       data=data)
        if not resp.is_ok:
            resp.data = None
            return resp
        response_data = resp.data
        response_xml_data = ET.fromstring(response_data)
        files_data = [File(single_file) for single_file in response_xml_data]
        resp.data = (files_data if not self.json_output
                     else [each.as_dict() for each in files_data])
        return resp

    def isfile(self, uid, path):
        """ Check file of given user exists.
        """
        dirname, basename = path.rsplit("/", 1)
        _dirs, files = self.lsdir(uid, dirname)
        return basename in (files or [])

    def isdir(self, uid, path):
        """ Check dir of given user exists.
        """
        dirname, basename = path.rsplit("/", 1)
        dirs, _files = self.lsdir(uid, dirname)
        return basename in (dirs or [])

    def lsdir(self, uid, path):
        """ List directory of a given user.
        """
        assert path == "" or path.startswith("/")
        resp = self.list_folders(uid, path=path, depth=1)
        dirs, files = [], []
        if resp.data is None:
            return None, None
        for item in resp.data[1:]:
            if item["resource_type"] == "collection":
                dirs.append(item["href"].split("/")[-2])
            elif item["resource_type"] is None:
                files.append(item["href"].split("/")[-1])
            else:
                raise ValueError("Unknown resource type!")
        return dirs, files

    def download_file(self, uid, path):
        """ Download file of given user by path.

        File will be saved to working directory path argument must be valid
        file path
        Exception will be raised if path doesn't exist, path is a directory,
        file with same name already exists in working directory.
        """
        additional_url = f"{uid}/{path}"
        filename = path.split('/')[-1] if '/' in path else path
        file_data = self.list_folders(uid=uid, path=path, depth=0)
        if not file_data:
            raise ValueError("Given path doesn't exist")
        file_resource_type = (file_data.data[0].get('resource_type')
                              if self.json_output
                              else file_data.data[0].resource_type)
        if file_resource_type == File.COLLECTION_RESOURCE_TYPE:
            raise ValueError("This is a collection, please specify file path")
        if filename in os.listdir('./'):
            raise ValueError(
                "File with such name already exists in this directory")
        res = self.requester.download(additional_url)
        with open(filename, 'wb') as f:
            f.write(res.data)

    def upload_file(self, uid, local_filepath, remote_filepath):
        """ Upload file to Nextcloud storage.
        """
        with open(local_filepath, 'rb') as f:
            file_content = f.read()
        additional_url = f"{uid}/{remote_filepath}"
        return self.requester.put(additional_url, data=file_content)

    def create_folder(self, uid, folder_path):
        """ Create folder on Nextcloud storage.
        """
        return self.requester.make_collection(
            additional_url=f"{uid}/{folder_path}")

    def delete_path(self, uid, path):
        """ Delete file or folder with all content of given user by path.
        """
        url = f"{uid}/{path}"
        return self.requester.delete(url=url)

    def move_path(self, uid, path, destination_path, overwrite=False):
        """ Move file or folder to destination.
        """
        path_url = f"{uid}/{path}"
        destination_path_url = f"{uid}/{destination_path}"
        return self.requester.move(
            url=path_url, destination=destination_path_url,
            overwrite=overwrite)

    def copy_path(self, uid, path, destination_path, overwrite=False):
        """ Copy file or folder to destination.
        """
        path_url = f"{uid}/{path}"
        destination_path_url = f"{uid}/{destination_path}"
        return self.requester.copy(
            url=path_url, destination=destination_path_url,
            overwrite=overwrite)

    def set_favorites(self, uid, path):
        """ Set files of a user favorite.
        """
        data = """<?xml version="1.0"?>
        <d:propertyupdate xmlns:d="DAV:" xmlns:oc="http://owncloud.org/ns">
          <d:set>
                <d:prop>
                  <oc:favorite>1</oc:favorite>
                </d:prop>
          </d:set>
        </d:propertyupdate>
        """
        url = f"{uid}/{path}"
        return self.requester.proppatch(additional_url=url, data=data)

    def list_favorites(self, uid, path=""):
        """ Set files of a user favorite.
        """
        data = """<?xml version="1.0"?>
        <oc:filter-files xmlns:d="DAV:"
                         xmlns:oc="http://owncloud.org/ns"
                         xmlns:nc="http://nextcloud.org/ns">
                 <oc:filter-rules>
                         <oc:favorite>1</oc:favorite>
                 </oc:filter-rules>
         </oc:filter-files>
        """
        url = f"{uid}/{path}"
        res = self.requester.report(additional_url=url, data=data)
        if not res.is_ok:
            res.data = None
            return res
        response_xml_data = ET.fromstring(res.data)
        files_data = [File(single_file) for single_file in response_xml_data]
        res.data = (files_data if not self.json_output
                    else [each.as_dict() for each in files_data])
        return res


class File:

    SUCCESS_STATUS = 'HTTP/1.1 200 OK'

    # key is NextCloud property, value is python variable name
    FILE_PROPERTIES = {
        # d:
        "getlastmodified": "last_modified",
        "getetag": "etag",
        "getcontenttype": "content_type",
        "resourcetype": "resource_type",
        "getcontentlength": "content_length",
        # oc:
        "id": "id",
        "fileid": "file_id",
        "favorite": "favorite",
        "comments-href": "comments_href",
        "comments-count": "comments_count",
        "comments-unread": "comments_unread",
        "owner-id": "owner_id",
        "owner-display-name": "owner_display_name",
        "share-types": "share_types",
        "checksums": "check_sums",
        "size": "size",
        "href": "href",
        # nc:
        "has-preview": "has_preview",
    }
    xml_namespaces_map = {
        "d": "DAV:",
        "oc": "http://owncloud.org/ns",
        "nc": "http://nextcloud.org/ns"
    }
    COLLECTION_RESOURCE_TYPE = 'collection'

    def __init__(self, xml_data):
        self.href = xml_data.find('d:href', self.xml_namespaces_map).text
        for propstat in xml_data.iter('{DAV:}propstat'):
            if (propstat.find('d:status', self.xml_namespaces_map).text !=
                    self.SUCCESS_STATUS):
                continue
            for file_property in propstat.find('d:prop',
                                               self.xml_namespaces_map):
                file_property_name = re.sub("{.*}", "", file_property.tag)
                if file_property_name not in self.FILE_PROPERTIES:
                    continue
                if file_property_name == 'resourcetype':
                    value = self._extract_resource_type(file_property)
                else:
                    value = file_property.text
                setattr(self, self.FILE_PROPERTIES[file_property_name], value)

    def _extract_resource_type(self, file_property):
        file_type = list(file_property)
        if file_type:
            return re.sub("{.*}", "", file_type[0].tag)
        return None

    def as_dict(self):
        return {key: value
                for key, value in self.__dict__.items()
                if key in self.FILE_PROPERTIES.values()}


class WebDAVStatusCodes:
    CREATED_CODE = 201
    NO_CONTENT_CODE = 204
    MULTISTATUS_CODE = 207
    ALREADY_EXISTS_CODE = 405
    PRECONDITION_FAILED_CODE = 412



##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module defines the DWI bvec/bval dataset loader.
"""

# Third party import
import numpy as np

# Package import
from .loader_base import LoaderBase


class DWI(LoaderBase):
    """ Define the DWI bvec bval loader.
    """
    allowed_extensions = [".bval", ".bvec"]

    def load(self, path):
        """ A method that load a bvec/bvel file.

        Parameters
        ----------
        path: str
            the path to the bvec/bval file.

        Returns
        -------
        data: object
            the loaded data.
        """
        return np.loadtxt(path)

    def save(self, data, outpath):
        """ A method that save the data in a bvec/bval file.

        Parameters
        ----------
        data: object
            the data to be saved.
        outpath: str
            the path where the the data will be saved.
        """
        np.savetxt(outpath, data)



##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module defines the JSON dataset loader.
"""

# Third party import
import json

# Package import
from .loader_base import LoaderBase


class JSON(LoaderBase):
    """ Define the JSON loader.
    """
    allowed_extensions = [".json", ]

    def load(self, path):
        """ A method that loads a JSON file.

        Parameters
        ----------
        path: str
            the path to the JSON file.

        Returns
        -------
        data: object
            the loaded data.
        """
        with open(path, "rt") as open_file:
            data = json.load(open_file)
        return data

    def save(self, data, path):
        """ A method that saves data into a JSON file.

        Parameters
        ----------
        data: object
            the data to be saved.
        path: str
            the path to the JSON file to save data into.
        """
        with open(path, "wt") as open_file:
            json.dump(data, open_file, indent=4)



###############################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
###############################################################################


"""
This module provides common utilities.
"""

# Imports
import os
import re
import json
import time
import functools
from docx import Document


def export_report(report, timestamp, outfile):
    """ Export the report in docx format.

    Parameters
    ----------
    report: str
        a dictionary with recursive keys and values.
    timestamp: str
        a timestamp.
    outfile: str
        the path to the generated docx file.
    """
    document = Document()
    section = document.sections[0]
    header = section.header
    paragraph = header.paragraphs[0]
    paragraph.text = f"NeuroSpin\tReporting\t{timestamp}"
    paragraph = document.add_paragraph(
        f"\n\n\n\nYou will find below the report generated on '{timestamp}'. "
        "If you have any questions please use the contact mail: rlink@cea.fr.")
    for family, family_item in report.items():
        document.add_heading(family.replace(".", " ").title())
        for validator, validator_item in family_item.items():
            split_validator = re.findall("[A-Z][^A-Z]*", validator)
            document.add_heading(" ".join(split_validator))
            paragraph = document.add_paragraph(
                "\n\n Below the table summarizing the errors.\n\n")
            for key, values in validator_item.items():
                table = document.add_table(rows=len(values), cols=2)
                cell = table.cell(0, 0)
                cell.text = key
                for idx, val in enumerate(values):
                    cell = table.cell(idx, 1)
                    cell.text = val
    document.save(outfile)


def monitor(func):
    """ A decorator to monitor function and log its status in a root directory.
    The input function parameters must be set via the 'CARAVEL_ROOT'
    and 'CARAVEL_NAME' environement variables.
    """
    root = os.environ.get("CARAVEL_ROOT", None)
    name = os.environ.get("CARAVEL_NAME", None)
    is_monitor = root is not None and name is not None
    if is_monitor:
        assert os.path.isdir(root), root

    @functools.wraps(func)
    def decorated(*args, **kwargs):
        try:
            tic = time.time()
            res = func(*args, **kwargs)
            toc = time.time()
            info = {
                "status": "healthy",
                "duration": toc - tic
            }
            if is_monitor:
                with open(os.path.join(root, f"{name}.json"), "wt") as of:
                    json.dump(info, of, indent=4)
            return res
        except Exception as e:
            info = {
                "status": "dead"
            }
            if is_monitor:
                with open(os.path.join(root, f"{name}.json"), "wt") as of:
                    json.dump(info, of, indent=4)
            raise e
    return decorated



from ..base import WithRequester


class User(WithRequester):
    API_URL = "/ocs/v1.php/cloud/users"
    SUCCESS_CODE = 100

    def add_user(self, uid, passwd):
        """ Create a new user on the Nextcloud server.
        """
        msg = {'userid': uid, 'password': passwd}
        return self.requester.post("", msg)

    def get_users(self, search=None, limit=None, offset=None):
        """ Retrieve a list of users from the Nextcloud server.
        """
        params = {
            'search': search,
            'limit': limit,
            'offset': offset
        }
        return self.requester.get(params=params)

    def get_user(self, uid):
        """ Retrieve information about a single user.
        """
        return self.requester.get(f"{uid}")

    def edit_user(self, uid, what, value):
        """ Edit attributes related to a user.

        Users are able to edit email, displayname and password; admins can
        also edit the quota value.
        """
        what_to_key_map = {
            "email": "email", "quota": "quota", "phone": "phone",
            "address": "address", "website": "website", "twitter": "twitter",
            "displayname": "displayname", "password": "password"
        }
        assert what in what_to_key_map, (
            "You have chosen to edit user's '{what}', but you can choose only"
            "from: {choices}".format(
                what=what, choices=", ".join(what_to_key_map.keys()))
        )
        url = f"{uid}"
        msg = {
            "key": what_to_key_map[what],
            "value": value,
        }
        return self.requester.put(url, msg)

    def disable_user(self, uid):
        """ Disable a user on the Nextcloud server so that the user cannot
        login anymore.
        """
        return self.requester.put(f"{uid}/disable")

    def enable_user(self, uid):
        """ Enable a user on the Nextcloud server so that the user can login
        again.
        """
        return self.requester.put(f"{uid}/enable")

    def delete_user(self, uid):
        """ Delete a user from the Nextcloud server.
        """
        return self.requester.delete(f"{uid}")

    def add_to_group(self, uid, gid):
        """ Add the specified user to the specified group.
        """
        url = f"{uid}/groups"
        msg = {'groupid': gid}
        return self.requester.post(url, msg)

    def remove_from_group(self, uid, gid):
        """ Remove the specified user from the specified group.
        """
        url = f"{uid}/groups"
        msg = {'groupid': gid}
        return self.requester.delete(url, msg)

    def create_subadmin(self, uid, gid):
        """ Make a user the subadmin of a group.
        """
        url = f"{uid}/subadmins"
        msg = {'groupid': gid}
        return self.requester.post(url, msg)

    def remove_subadmin(self, uid, gid):
        """ Remove the subadmin rights for the user specified from the group
        specified.
        """
        url = f"{uid}/subadmins"
        msg = {'groupid': gid}
        return self.requester.delete(url, msg)

    def get_subadmin_groups(self, uid):
        """ Get the groups in which the user is a subadmin.
        """
        url = f"{uid}/subadmins"
        return self.requester.get(url)

    def resend_welcome_mail(self, uid):
        """ Trigger the welcome email for this user again.
        """
        url = f"{uid}/welcome"
        return self.requester.post(url)



##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module contains generic functions to load/save a dataset.
"""

# Package import
from caravel.loaders import TSV
from caravel.loaders import NIFTI
from caravel.loaders import JSON
from caravel.loaders import DWI
from caravel.loaders import TARGZ
from caravel.loaders import PNG
from caravel.loaders import CSV
from caravel.loaders import XLSX
from caravel.loaders import PDF
from caravel.loaders import PLINK
from caravel.loaders import VCF
from caravel.loaders import MP4
from caravel.loaders import EDF
from caravel.loaders import MZML

# Global parameters
# > define all the available loaders
LOADERS = [MP4, EDF, PNG, TSV, CSV, NIFTI, JSON, DWI, TARGZ, XLSX, PDF,
           PLINK, VCF, MZML]


def load(path, **kwargs):
    """ Load the data.

    Parameters
    ----------
    path: str
        the path to the data to be loaded.

    Returns
    -------
    data: object
        the loaded data.
    """
    loader = get_loader(path)
    return loader.load(path, **kwargs)


def save(data, path, **kwargs):
    """ Save the data.

    Parameters
    ----------
    data: object
        the data to be saved.
    path: str
        the destination file.
    """
    saver = get_saver(path)
    saver.save(data, path, **kwargs)


def get_loader(path):
    """ Search for a suitable loader in the declared loaders.
    Raise an exception if no loader is found.

    Parameters
    ----------
    path: str
        the path to the data to be loaded.

    Returns
    -------
    loader: @instance
        the loader instance.
    """
    for loader_class in LOADERS:
        loader = loader_class()
        if loader.can_load(path):
            return loader
    raise Exception(f"No loader available for '{path}'.")


def get_saver(path):
    """ Search for a suitable saver in the declared savers.
    Raise an exception if no saver is found.

    Parameters
    ----------
    path: str
        the path to the data to be saved.

    Returns
    -------
    saver: @instance
        the loader instance.
    """
    for saver_class in LOADERS:
        saver = saver_class()
        if saver.can_save(path):
            return saver
    raise Exception(f"No saver available for '{path}'.")



##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module contains the generic parser definition.
"""

# System import
import os
import json
import glob
import pickle
import datetime

# Third party import
import pandas as pd

# Package import
from caravel.io import load


class ParserBase:
    """ Base parser to retrieve data from a BIDS directory.
    """
    AVAILABLE_LAYOUTS = ("sourcedata", "rawdata", "derivatives", "phenotype")

    def __init__(self, project, confdir, layoutdir):
        """ Initialize the Caravel class.

        Parameters
        ----------
        project: str
            the name of the project you are working on.
        confdir: str
            the locations of the configuration file of the current project.
        layoutdir: str
            the location of the pre-generated parsing representations. If None
            switch to managers mode.
        """
        self.project = project
        self.layouts = {}
        _conf = ParserBase._get_conf(confdir)
        if project not in _conf:
            raise ValueError(
                "Unknown configuration for project '{}'. Available projects "
                "are: {}.".format(project, _conf.keys()))
        self.conf = _conf[project]
        if layoutdir is not None:
            _repr = self._get_repr(layoutdir)
            if project not in _repr:
                raise ValueError(
                    f"Unknown representation for project '{project}'. "
                    f"Available projects are: {_repr.keys()}.")
            self.representation = _repr[project]
        else:
            self.representation = {"manager": [{"path": "to_be_created.pkl"}]}
        self.connection = None

    def can_load(self):
        """ A method checking the dataset type.

        Returns
        -------
        out: bool
            True if the dataset can be loaded, False otherwise.
        """
        checks = [elem[-1]["path"] for elem in self.representation.values()]
        if len(checks) == 0:
            return False
        return all(elem.endswith(self.EXT) for elem in checks)

    def _check_layout(self, name):
        """ Check if the layout name is supported.
        """
        if name not in self.AVAILABLE_LAYOUTS:
            raise ValueError(
                f"Layout '{name}' is not yet supported. "
                f"Available layouts are: {self.AVAILABLE_LAYOUTS}.")

    @classmethod
    def _get_conf(cls, confdir):
        """ List all the configurations available and sort them by project.
        """
        conf = {}
        for path in glob.glob(os.path.join(confdir, "*.conf")):
            basename = os.path.basename(path).replace(".conf", "")
            project, name = basename.split("_")
            if project not in conf:
                conf[project] = {}
            conf[project][name] = path
        return conf

    def _get_repr(self, layoutdir):
        """ List all the layout representation available and sort them by
        dates.
        """
        representations = {}
        layout_files = glob.glob(os.path.join(layoutdir, "*.pkl"))
        for path in layout_files:
            basename = os.path.basename(path).replace(".pkl", "")
            project, name, timestamp = basename.split("_")
            if project not in representations:
                representations[project] = {}
            representations[project].setdefault(name, []).append(
                {"date": timestamp, "path": path})
        for project, project_data in representations.items():
            for name, name_data in project_data.items():
                name_data.sort(key=lambda x: datetime.datetime.strptime(
                    x["date"], "%Y-%m-%d"))
        return representations

    def _check_conf(self, name):
        """ Check if configuration is declared for the layout.
        """
        if name not in self.conf:
            raise ValueError(
                "No configuration available for layout '{0}'. Please contact "
                "the module developers to add the support for your project.")

    def _load_layout(self, name):
        """ Load a layout from its pre-generated representation.
        """
        if name not in self.layouts:
            if name not in self.representation:
                raise ValueError(
                    f"A pre-generated '{name}' layout for your project "
                    f"'{self.project}' is expected in user mode. Please "
                    "contact the developers of the module.")
            path = self.representation[name][-1]["path"]
            with open(path, "rb") as open_file:
                self.layouts[name] = pickle.load(open_file)
        return self.layouts[name]

    def _load_conf(self, name):
        """ Load the configuration associated to a layout.
        """
        if not isinstance(self.conf[name], dict):
            with open(self.conf[name], "rt") as open_file:
                self.conf[name] = json.load(open_file)

    def export_layout(self, name):
        """ Export a layout as a pandas DataFrame.

        Parameters
        ----------
        name: str
            the name of the layout.

        Returns
        -------
        df: pandas DataFrame
            the converted layout.
        """
        raise NotImplementedError("This function has to be defined in child "
                                  "child class.")

    def list_keys(self, name):
        """ List all the filtering keys available in the layout.

        Parameters
        ----------
        name: str
            the name of the layout.

        Returns
        -------
        keys: list
            the layout keys.
        """
        raise NotImplementedError("This function has to be defined in child "
                                  "child class.")

    def list_values(self, name, key):
        """ List all the filtering key values available in the layout.

        Parameters
        ----------
        name: str
            the name of the layout.
        key: str
            the name of key in the layout.

        Returns
        -------
        values: list
            the key associated values in the layout.
        """
        raise NotImplementedError("This function has to be defined in child "
                                  "child class.")

    def filter_layout(self, name, extensions=None, **kwargs):
        """ Filter the layout by using a combination of key-values rules.

        Parameters
        ----------
        name: str
            the name of the layout.
        extensions: str or list of str
            a filtering rule on the file extension.
        kwargs: dict
            the filtering options.

        Returns
        -------
        df: pandas DataFrame
            the filtered layout.
        """
        raise NotImplementedError("This function has to be defined in child "
                                  "child class.")

    def load_data(self, name, df, replace=None):
        """ Load the data contained in the filename column of a pandas
        DataFrame.

        Note:
        Only a couple of file extensions are supported. If no loader has been
        found None is returned.

        Parameters
        ----------
        name: str
            the name of the layout.
        df: pandas DataFrame
            a table with one 'filename' column.
        replace: 2-uplet, default None
            in the case of a CubicWeb resource, the data are downloaded in a
            custom folder. Use this parameter to replace the server location
            by your own location.

        Returns
        -------
        data: dict
            a dictionaray containing the loaded data.
        """
        if "filename" not in df:
            raise ValueError("One 'filename' column expected in your table.")
        data = {}
        for index, path in enumerate(df["filename"]):
            if isinstance(path, dict):
                _data = pd.DataFrame.from_records([path])
                path = [f"{key}-{val}"
                        for key, val in zip(df.columns, df.to_numpy()[index])
                        if key != "filename"]
                path = "_".join(path)
            else:
                if replace is not None:
                    path = path.replace(replace[0], replace[1])
                try:
                    _data = load(path)
                except Exception:
                    _data = None
                if isinstance(_data, pd.DataFrame):
                    layout = self._load_layout(name)
                    file_obj = layout.files[path]
                    for ent_name, ent_val in file_obj.entities.items():
                        if ent_name in self.BASE_ENTITIES:
                            _data[ent_name] = ent_val
                    _data["dtype"] = name
                    if "participant_id" in _data:
                        _data["participant_id"] = _data[
                            "participant_id"].str.replace("sub-", "")
            data[path] = _data
        return data



##########################################################################
# NSAp - Copyright (C) CEA, 2020
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module defines the edf dataset loader.
"""

# Third party import
import pyedflib

# Package import
from .loader_base import LoaderBase


class EDF(LoaderBase):
    """ Define the mp4 loader.
    """
    allowed_extensions = [".edf"]

    def load(self, path):
        """ A method that load the edf data.

        Parameters
        ----------
        path: str
            the path to the edf to be loaded.

        Returns
        -------
        data: pyedflib object
        """
        return pyedflib.EdfReader(path)

    def save(self, path, signals, signal_headers, header=None):
        """ A method that save the image in edf.
        """
        pyedflib.highlevel.write_edf(path,
                                     signals=signals,
                                     signal_headers=signal_headers,
                                     header=header)



##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module defines the CSV dataset loader.
"""

# Third party import
from pandas_plink import read_plink1_bin

# Package import
from .loader_base import LoaderBase


class PLINK(LoaderBase):
    """ Define the TSV loader.
    """
    allowed_extensions = [".bed"]

    def load(self, path):
        """ A method that load the table data.

        Parameters
        ----------
        path: str
            the path to the table to be loaded.

        Returns
        -------
        data: xarray.core.dataarray.DataArray
            the loaded table.
        """
        G = read_plink1_bin(path, verbose=False)

        return G



##########################################################################
# NSAp - Copyright (C) CEA, 2020
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module defines the mp4 dataset loader.
"""

# Third party import
import imageio

# Package import
from .loader_base import LoaderBase


class MP4(LoaderBase):
    """ Define the mp4 loader.
    """
    allowed_extensions = [".mp4"]

    def load(self, path):
        """ A method that load the mp4 data.

        Parameters
        ----------
        path: str
            the path to the mp4 file to be loaded.

        Returns
        -------
        data: imageio numpy array
            the loaded image.
        """
        return imageio.get_reader(path,  'ffmpeg')

    def save(self, data, outpath, fps=24):
        """ A method that save the image in mp4.

        Parameters
        ----------
        data: list of path
            list of path for each image for the video.
        outpath: str
            the path where the the mp4 image will be saved.
        """

        writer = imageio.get_writer(outpath, fps)
        for png_path in data:
            im = imageio.imread(png_path),
            writer.append_data(im[:, :, 1])
        writer.close()



##########################################################################
# NSAp - Copyright (C) CEA, 2019 - 2024
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

# Module current version
version_major = 1
version_minor = 0
version_micro = 0

# Expected by setup.py: string of form "X.Y.Z"
__version__ = f"{version_major}.{version_minor}.{version_micro}"

# Expected by setup.py: the status of the project
CLASSIFIERS = ["Development Status :: 1 - Planning",
               "Environment :: Console",
               "Operating System :: OS Independent",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering"]

# Project descriptions
description = """
pycaravel:  This module is designed to simplify continuous integration (CI) of
data from multiple projects.
"""
SUMMARY = """
.. container:: summary-carousel

    pycaravel is a Python module to simplify continuous integration (CI) of
    data from multiple projects:

    1. a common API for parsing multiple source of data (currently only BIDS).
    2. a common API to search in those datasets.
    3. a common API to validate incoming data.
"""
long_description = (
    "This module is designed to simplify continuous integration (CI) "
    "of data from multiple projects. ")

# Main setup parameters
NAME = "pycaravel"
ORGANISATION = "CEA"
MAINTAINER = "Antoine Grigis"
MAINTAINER_EMAIL = "antoine.grigis@cea.fr"
DESCRIPTION = description
LONG_DESCRIPTION = long_description
EXTRANAME = "NeuroSpin webPage"
EXTRAURL = ("http://joliot.cea.fr/drf/joliot/Pages/Entites_de_recherche/"
            "NeuroSpin.aspx")
URL = "https://github.com/neurospincloud-ci/pycaravel"
DOWNLOAD_URL = "https://github.com/neurospincloud-ci/pycaravel"
LICENSE = "CeCILL-B"
CLASSIFIERS = CLASSIFIERS
AUTHOR = """
Antoine Grigis <antoine.grigis@cea.fr>
"""
AUTHOR_EMAIL = "antoine.grigis@cea.fr"
PLATFORMS = "Linux,OSX"
ISRELEASE = True
VERSION = __version__
PROVIDES = ["caravel"]
REQUIRES = [
    "pandas",
    "openpyxl",
    "grabbit @ git+https://github.com/grabbles/grabbit.git",
    "nibabel",
    "numpy",
    "imageio[ffmpeg]",
    "PyPDF2",
    "vcfpy",
    "pandas-plink",
    "pyEDFlib",
    "requests",
    "python-docx",
    "fire",
    "pymzml"
]
EXTRA_REQUIRES = {}
SCRIPTS = [
    "caravel/scripts/project-ci"
]



"""
NextCloud Python API from https://github.com/matejak/nextcloud-API.git.
"""

from .activity import Activity
from .apps import Apps
from .capabilities import Capabilities
from .federated_cloudshares import FederatedCloudShare
from .group import Group
from .group_folders import GroupFolders
from .notifications import Notifications
from .share import Share
from .user import User
from .user_ldap import UserLDAP
from .webdav import WebDAV
from .forms import Forms


OCS_API_CLASSES = [Activity, Apps, Capabilities, FederatedCloudShare, Group,
                   GroupFolders, Notifications, Share, User, UserLDAP, Forms]
WEBDAV_CLASS = WebDAV



from ..base import WithRequester


class GroupFolders(WithRequester):
    API_URL = "/apps/groupfolders/folders"
    SUCCESS_CODE = 100

    def get_group_folders(self):
        """ Return a list of call configured folders and their settings.
        """
        return self.requester.get()

    def get_group_folder(self, fid):
        """ Return a specific configured folder and it's settings.
        """
        return self.requester.get(fid)

    def create_group_folder(self, mountpoint):
        """ Create a new group folder.
        """
        return self.requester.post(data={"mountpoint": mountpoint})

    def delete_group_folder(self, fid):
        """ Delete a group folder.
        """
        return self.requester.delete(fid)

    def grant_access_to_group_folder(self, fid, gid):
        """ Give a group access to a folder.
        """
        url = "/".join([str(fid), "groups"])
        return self.requester.post(url, data={"group": gid})

    def revoke_access_to_group_folder(self, fid, gid):
        """ Remove access from a group to a folder.
        """
        url = "/".join([str(fid), "groups", gid])
        return self.requester.delete(url)

    def set_permissions_to_group_folder(self, fid, gid, permissions):
        """ Set the permissions a group has in a folder.
        """
        url = "/".join([str(fid), "groups", gid])
        return self.requester.post(url=url, data={"permissions": permissions})

    def set_quota_of_group_folder(self, fid, quota):
        """ Set the quota for a folder in bytes.
        """
        url = "/".join([str(fid), "quota"])
        return self.requester.post(url, {"quota": quota})

    def rename_group_folder(self, fid, mountpoint):
        """ Change the name of a folder.
        """
        url = "/".join([str(fid), "mountpoint"])
        return self.requester.post(url=url, data={"mountpoint": mountpoint})



##########################################################################
# NSAp - Copyright (C) CEA, 2020
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
A module to simplify email sending.
"""


# System import
import os
from email import encoders
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
import smtplib
import mimetypes


class EmailManager:
    """ Define a class to simplify emails sending.
    """
    def __init__(self, smtp_host, smtp_port):
        """ Initialize the EmailManager class.

        Parameters
        ----------
        smtp_host: str
            the SMTP host.
        smtp_port: str
            the SMTP port.
        """
        self.host = smtp_host
        self.port = smtp_port

    def send_mail(self, to_addrs, subject, body, from_addr="noreply@cea.fr",
                  files=None):
        """ Send an email.

        Parameters
        ----------
        to_addrs: list of str
            the destination emails
        subject: str
            the subject of the email.
        body: str
            the body of the email.
        from_addr: str, default 'noreply@cea.fr'
            the sender email.
        files: list of str default None
            a list of files to be attached.
        """
        # Create the container with the email message.
        msg = MIMEMultipart()
        msg["Subject"] = subject
        msg["From"] = from_addr
        msg["To"] = ", ".join(to_addrs)
        msg.attach(MIMEText(body, "plain", "utf-8"))
        files = files or []
        for path in files:
            attachment = self._create_attachment(path)
            msg.attach(attachment)

        # Send the email via our own SMTP server.
        with smtplib.SMTP(self.host, self.port) as server:
            server.send_message(msg)

    def _create_attachment(self, path):
        ctype, encoding = mimetypes.guess_type(path)
        if ctype is None or encoding is not None:
            ctype = "application/octet-stream"
        maintype, subtype = ctype.split("/", 1)
        with open(path, "rb") as fp:
            attachment = MIMEBase(maintype, subtype)
            attachment.set_payload(fp.read())
        encoders.encode_base64(attachment)
        filename = os.path.basename(path)
        attachment.add_header(
            "Content-Disposition", "attachment", filename=filename
        )
        return attachment



from json import JSONDecodeError
from .api_wrappers.webdav import WebDAVStatusCodes


class NextCloudResponse:
    def __init__(self, response, json_output=True, data=None):
        self.raw = response
        if not data:
            self.data = (
                response.json() if json_output
                else response.content.decode("UTF-8"))
        else:
            self.data = data


class WebDAVResponse(NextCloudResponse):
    """ Response class for WebDAV api methods.
    """
    METHODS_SUCCESS_CODES = {
        "PROPFIND": [WebDAVStatusCodes.MULTISTATUS_CODE],
        "PROPPATCH": [WebDAVStatusCodes.MULTISTATUS_CODE],
        "REPORT": [WebDAVStatusCodes.MULTISTATUS_CODE],
        "MKCOL": [WebDAVStatusCodes.CREATED_CODE],
        "COPY": [WebDAVStatusCodes.CREATED_CODE,
                 WebDAVStatusCodes.NO_CONTENT_CODE],
        "MOVE": [WebDAVStatusCodes.CREATED_CODE,
                 WebDAVStatusCodes.NO_CONTENT_CODE],
        "PUT": [WebDAVStatusCodes.CREATED_CODE],
        "DELETE": [WebDAVStatusCodes.NO_CONTENT_CODE]
    }

    def __init__(self, response, data=None):
        super().__init__(
            response=response, data=data, json_output=False)
        request_method = response.request.method
        self.is_ok = False
        if request_method in self.METHODS_SUCCESS_CODES:
            self.is_ok = response.status_code in self.METHODS_SUCCESS_CODES[
                request_method]

    def __repr__(self):
        is_ok_str = "OK" if self.is_ok else "Failed"
        return f"<OCSResponse: Status: {is_ok_str}>"


class OCSResponse(NextCloudResponse):
    """ Response class for OCS api methods.
    """
    def __init__(self, response, json_output=True, success_code=None):
        self.raw = response
        self.is_ok = None

        if json_output:
            try:
                self.full_data = response.json()
                self.meta = self.full_data['ocs']['meta']
                self.status_code = self.full_data['ocs']['meta']['statuscode']
                self.data = self.full_data['ocs']['data']
                if success_code:
                    self.is_ok = (
                        self.full_data['ocs']['meta']['statuscode'] ==
                        success_code)
            except JSONDecodeError:
                self.is_ok = False
                self.data = {'message': 'Unable to parse JSON response'}
        else:
            self.data = response.content.decode("UTF-8")

    def __repr__(self):
        is_ok_str = "OK" if self.is_ok else "Failed"
        return f"<OCSResponse: Status: {is_ok_str}>"



###############################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
###############################################################################


"""
This module provides a registry with function to be applied during
the validation stage.
"""

# Imports
import re
import logging
import warnings
import traceback
from operator import itemgetter
from collections import namedtuple
from .info import __version__
logger = logging.getLogger("caravel")


class MetaRegister(type):
    """ Simple Python metaclass registry pattern.
    """
    REGISTRY = {}

    def __new__(cls, name, bases, attrs):
        """ Allocation.

        Parameters
        ----------
        name: str
            the name of the class.
        bases: tuple
            the base classes.
        attrs:
            the attributes defined for the class.
        """
        new_cls = type.__new__(cls, name, bases, attrs)
        if name in cls.REGISTRY:
            raise ValueError(
                f"'{name}' name already used in registry.")
        if name != "ValidationBase":
            cls.REGISTRY[name] = new_cls
        return new_cls


class ValidationBase(metaclass=MetaRegister):
    """ A validation test must inherit from this base class.
    """
    LEVELS = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL
    }
    __family__ = "default"
    __priority__ = 7
    __function__ = None
    __level__ = "error"

    def __init__(self):
        """ Initialize the ValidationBase class.
        """
        ValidationBase.setup_logging()

    @classmethod
    def setup_logging(cls, logfile=None):
        """ Setup the logging.

        Parameters
        ----------
        logfile: str, default None
            the log file.
        """
        logging_format = logging.Formatter(
            "[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - "
            "%(message)s", "%Y-%m-%d %H:%M:%S")
        while len(logging.root.handlers) > 0:
            logging.root.removeHandler(logging.root.handlers[-1])
        while len(logger.handlers) > 0:
            logger.removeHandler(logger.handlers[-1])
        level = cls.LEVELS.get(cls.__level__, None)
        if level is None:
            raise ValueError("Unknown logging level.")
        logger.setLevel(level)
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(level)
        stream_handler.setFormatter(logging_format)
        logger.addHandler(stream_handler)
        if logfile is not None:
            file_handler = logging.FileHandler(logfile, mode="a")
            file_handler.setLevel(level)
            file_handler.setFormatter(logging_format)
            logger.addHandler(file_handler)
        if level != logging.DEBUG:
            warnings.simplefilter("ignore", DeprecationWarning)

    def __call__(self, data):
        """ The method to run the test.

        Parameters
        ----------
        data: dict
            the input data for the test.
        """
        logger.info(f"running '{self.__function__.__name__}' "
                    f"with version '{__version__}' of pycaravel")
        return self.__function__(**data)


class ValidationDecorator:
    """ Dynamically create a validator.

    In order to make the class publicly accessible, we assign the result of
    the function to a variable dynamically using globals().
    """
    def __init__(self, family="default", priority=7):
        """ Initialize the ValidationDecorator class.

        Parameters
        ----------
        family: str, default 'default'
            the family name of the test.
        priority: int, default 7
            the prioprity run of the validation.
        """
        self.destination_module_globals = globals()
        self.family = family
        self.priority = priority

    def __call__(self, function, *args, **kwargs):
        """ Create the validator.

        Parameters
        ----------
        function: callable
            the function that perform the test.
        """
        class_name = function.__name__.replace(
            "_", " ").title().replace(" ", "")
        mod_name = self.destination_module_globals["__name__"]
        class_parameters = {
            "__module__": mod_name,
            "_id":  mod_name + "." + class_name,
            "__function__": function,
            "__family__": self.family,
            "__priority__": self.priority
        }
        self.destination_module_globals[class_name] = (
            type(class_name, (ValidationBase, ), class_parameters))


def get_validators(family=None):
    """ List/sort all available validators.

    Parameters
    ----------
    family: str or list of str, default None
        the validators family name.

    Returns
    -------
    validators: dict
        the requested validators.
    """
    available_validators = sorted(ValidationBase.REGISTRY.keys())
    if family is not None and not isinstance(family, list):
        family = [family]
    validators = {}
    for key in available_validators:
        fct = ValidationBase.REGISTRY[key]
        if family is not None:
            for cnt, regex in enumerate(family):
                if re.match(regex, fct.__family__) is not None:
                    break
                cnt += 1
            if cnt == len(family):
                continue
        validators.setdefault(fct.__family__, []).append(fct())
    return validators


def listify(validators):
    """ Sort the validators by priority level.

    Parameters
    ----------
    validators: dict
        the validators as returned by the 'get_validators' function.

    Returns
    -------
    sorted_validators: list
        the validators sorted by priority level.
    """
    Validator = namedtuple("Validator", "name priority instance")
    sorted_validators = []
    for key, instances in validators.items():
        sorted_validators.extend([
            Validator(name=key, priority=instance.__priority__,
                      instance=instance) for instance in instances])
    sorted_validators = sorted(
        sorted_validators,
        key=itemgetter(Validator._fields.index("priority")))
    return sorted_validators


def run_validation(data, validators=None, logfile=None):
    """ Safely run a validation plane.

    Parameters
    ----------
    data: dict
        the validation data.
    validators: dict, default None
        the validators as returned by the 'get_validators' function.
    logfile: str, default None
        the log file.

    Returns
    -------
    report: dict
        the validation report.
    """
    ValidationBase.setup_logging(logfile=logfile)
    if validators is None:
        validators = get_validators()
    report = {}
    for validator in listify(validators):
        if validator.name not in report:
            report[validator.name] = {}
        try:
            result, extra_data = validator.instance(data)
            if isinstance(extra_data, dict):
                logger.info(f"adding extra kwargs '{extra_data.keys()}'")
                data.update(extra_data)
        except Exception:
            result = {"Internal error:": [traceback.format_exc()]}
        if result is None or result == {} or result == []:
            continue
        report[validator.name][validator.instance.__class__.__name__] = result
    return report



##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module contains the BIDS parser definition.
"""

# System import
import os
import pickle
import datetime

# Third party import
import numpy as np
import pandas as pd
from grabbit import Layout

# Package import
from .parser_base import ParserBase


class BIDSParser(ParserBase):
    """ Object to retrieve data from a BIDS directory.
    """
    BASE_ENTITIES = ["subject", "session", "task", "run", "suffix"]
    EXT = ".pkl"

    def export_layout(self, name):
        """ Export a layout as a pandas DataFrame.

        Parameters
        ----------
        name: str
            the name of the layout.

        Returns
        -------
        df: pandas DataFrame
            the converted layout.
        """
        layout = self._load_layout(name)
        return layout.as_data_frame()

    def list_keys(self, name):
        """ List all the filtering keys available in the layout.

        Parameters
        ----------
        name: str
            the name of the layout.

        Returns
        -------
        keys: list
            the layout keys.
        """
        layout = self._load_layout(name)
        return [elem.replace(f"{name}.", "")
                for elem in layout.entities]

    def list_values(self, name, key):
        """ List all the filtering key values available in the layout.

        Parameters
        ----------
        name: str
            the name of the layout.
        key: str
            the name of key in the layout.

        Returns
        -------
        values: list
            the key associated values in the layout.
        """
        layout = self._load_layout(name)
        _key = f"{name}.{key}"
        if _key not in layout.entities:
            raise ValueError(f"Unrecognize layout key '{key}'.")
        return list(layout.unique(_key))

    def filter_layout(self, name, extension=None, **kwargs):
        """ Filter the layout by using a combination of key-values rules.

        Parameters
        ----------
        name: str
            the name of the layout.
        extension: str or list of str
            a filtering rule on the file extension.
        kwargs: dict
            the filtering options.

        Returns
        -------
        df: pandas DataFrame
            the filtered layout.
        """
        layout = self._load_layout(name)
        if extension is not None:
            kwargs["extensions"] = extension
        header = None
        files = layout.get(**kwargs)
        if len(files) == 0:
            df = pd.DataFrame()
        else:
            file_obj = files[0]
            header = ["filename"] + self.list_keys(name)
            data = []
            for file_obj in files:
                row = [getattr(file_obj, key, np.nan) for key in header]
                data.append(row)
            df = pd.DataFrame(data, columns=header)
        df = df.dropna(axis="columns", how="all")
        return df

    def pickling_layout(self, bids_root, name, outdir, subset=None):
        """ Load the requested BIDS layout and save it as a pickle.

        Parameters
        ----------
        bids_root: str
            path to the BIDS folder.
        name: str
            the name of subfolder to be parsed (the layout name).
        outdir: str
            the folder where the pickle will be generated.
        subset: list of str, default None
            a selector to focus only on specific folders.

        Returns
        -------
        outfile: str
            the generated layout representation location.
        """
        self._check_layout(name)
        self._check_conf(name)
        layout_root = os.path.join(bids_root, name)
        if not os.path.isdir(layout_root):
            raise ValueError(f"'{layout_root}' is not a valid directory.")
        if subset is None:
            layout = Layout([(layout_root, self.conf[name])])
        else:
            layout = Layout([
                (os.path.join(layout_root, dirname), self.conf[name])
                for dirname in subset])
        self.layouts[name] = layout
        now = datetime.datetime.now()
        timestamp = f"{now.year}-{now.month}-{now.day}"
        outfile = os.path.join(
            outdir, f"{self.project}_{name}_{timestamp}.pkl")
        with open(outfile, "wb") as open_file:
            pickle.dump(layout, open_file, -1)
        return outfile



import enum


class WithRequester:
    API_URL = NotImplementedError

    def __init__(self, requester):
        self._requester = requester

    @property
    def requester(self):
        """ Get requester instance """
        # dynamically set API_URL for requester
        self._requester.API_URL = self.API_URL
        self._requester.SUCCESS_CODE = getattr(self, 'SUCCESS_CODE', None)
        return self._requester


class OCSCode(enum.IntEnum):
    OK = 100
    SERVER_ERROR = 996
    NOT_AUTHORIZED = 997
    NOT_FOUND = 998
    UNKNOWN_ERROR = 999


class ShareType(enum.IntEnum):
    USER = 0
    GROUP = 1
    PUBLIC_LINK = 3
    FEDERATED_CLOUD_SHARE = 6
    CIRCLE = 7


class Permission(enum.IntEnum):
    """ Permission for Share have to be sum of selected permissions.
    """
    READ = 1
    UPDATE = 2
    CREATE = 4
    DELETE = 8
    SHARE = 16
    ALL = 31


QUOTA_UNLIMITED = -3


def datetime_to_expire_date(date):
    return date.strftime("%Y-%m-%d")



from ..base import WithRequester


class Activity(WithRequester):
    API_URL = "/ocs/v2.php/apps/activity/api/v2/activity"
    SUCCESS_CODE = 200

    def get_activities(self, since=None, limit=None, object_type=None,
                       object_id=None, sort=None):
        """ Get an activity feed showing your file changes and other
        interesting things going on in your Nextcloud
        """
        params = {
            "since": since,
            "limit": limit,
            "object_type": object_type,
            "object_id": object_id,
            "sort": sort
        }
        if params['object_type'] and params['object_id']:
            return self.requester.get(url="filter", params=params)
        return self.requester.get(params=params)



import re
from ..base import WithRequester


class UserLDAP(WithRequester):
    API_URL = "/ocs/v2.php/apps/user_ldap/api/v1/config"
    SUCCESS_CODE = 200

    CONFIG_KEYS = [
        "ldapHost",
        "ldapPort",
        "ldapBackupHost",
        "ldapBackupPort",
        "ldapBase",
        "ldapBaseUsers",
        "ldapBaseGroups",
        "ldapAgentName",
        "ldapAgentPassword",
        "ldapTLS",
        "turnOffCertCheck",
        "ldapUserDisplayName",
        "ldapGidNumber",
        "ldapUserFilterObjectclass",
        "ldapUserFilterGroups",
        "ldapUserFilter",
        "ldapUserFilterMode",
        "ldapGroupFilter",
        "ldapGroupFilterMode",
        "ldapGroupFilterObjectclass",
        "ldapGroupFilterGroups",
        "ldapGroupMemberAssocAttr",
        "ldapGroupDisplayName",
        "ldapLoginFilter",
        "ldapLoginFilterMode",
        "ldapLoginFilterEmail",
        "ldapLoginFilterUsername",
        "ldapLoginFilterAttributes",
        "ldapQuotaAttribute",
        "ldapQuotaDefault",
        "ldapEmailAttribute",
        "ldapCacheTTL",
        "ldapUuidUserAttribute",
        "ldapUuidGroupAttribute",
        "ldapOverrideMainServer",
        "ldapConfigurationActive",
        "ldapAttributesForUserSearch",
        "ldapAttributesForGroupSearch",
        "ldapExperiencedAdmin",
        "homeFolderNamingRule",
        "hasPagedResultSupport",
        "hasMemberOfFilterSupport",
        "useMemberOfToDetectMembership",
        "ldapExpertUsernameAttr",
        "ldapExpertUUIDUserAttr",
        "ldapExpertUUIDGroupAttr",
        "lastJpegPhotoLookup",
        "ldapNestedGroups",
        "ldapPagingSize",
        "turnOnPasswordChange",
        "ldapDynamicGroupMemberURL",
        "ldapDefaultPPolicyDN",
    ]

    def create_ldap_config(self):
        """ Create a new and empty LDAP configuration.
        """
        return self.requester.post()

    def get_ldap_config(self, config_id, show_password=None):
        """ Get all keys and values of the specified LDAP configuration.
        """
        params = {"showPassword": show_password}
        return self.requester.get(config_id, params=params)

    def edit_ldap_config(self, config_id, data):
        """ Update a configuration with the provided values.
        """
        prepared_data = {f'configData[{key}]': value
                         for key, value in data.items()}
        return self.requester.put(config_id, data=prepared_data)

    def delete_ldap_config(self, config_id):
        """ Delete a given LDAP configuration.
        """
        return self.requester.delete(config_id)


for ldap_key in UserLDAP.CONFIG_KEYS:
    key_name = re.sub('ldap', '', ldap_key)
    key_name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', key_name).lower()

    # create and add getter method
    getter_name = f"get_ldap_{key_name}"

    def getter_method(param):
        def getter(self, config_id):
            res = self.get_ldap_config(config_id)
            data = res.data
            return data[param]
        getter.__name__ = getter_name
        return getter

    setattr(UserLDAP, getter_name, getter_method(ldap_key))

    # create and add setter method
    setter_name = f"set_ldap_{key_name}"

    def setter_method(param):
        def setter(self, config_id, value):
            res = self.edit_ldap_config(config_id, data={param: value})
            return res
        setter.__name__ = setter_name
        return setter

    setattr(UserLDAP, setter_name, setter_method(ldap_key))



##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module defines the loader base class.
"""


class LoaderBase:
    """ Base class for all loaders.
    """
    allowed_extensions = []

    def can_load(self, path):
        """ A method checking the file extension.

        Parameters
        ----------
        path: str
            the path to the data to be loaded.

        Returns
        -------
        out: bool
            True if the file extension is valid, False otherwise.
        """
        return any(path.endswith(ext) for ext in self.allowed_extensions)

    def load(self, path):
        """ A method that load the data and associated metadata.

        Parameters
        ----------
        path: str
            the path to the data to be loaded.

        Returns
        -------
        image: Data
            the loaded data.
        """
        raise NotImplementedError(
            "The 'load' method must be implemented in subclasses.")

    def can_save(self, outpath):
        """ A method checking the output file extension.

        Parameters
        ----------
        outpath: str
            the path where the the image will be saved.

        Returns
        -------
        out: bool
            True if the output file extension is valid, False otherwise.
        """
        return any(outpath.endswith(ext) for ext in self.allowed_extensions)

    def save(self, data, outpath):
        """ A method that save the data and associated metadata.

        Parameters
        ----------
        image: Data
            the data to be saved.
        outpath: str
            the path where the the data will be saved.
        """
        raise NotImplementedError(
            "The 'save' method must be implemented in subclasses.")



from ..base import WithRequester


class Apps(WithRequester):
    API_URL = "/ocs/v1.php/cloud/apps"
    SUCCESS_CODE = 100

    def get_apps(self, filter=None):
        """ Get a list of apps installed on the Nextcloud server.
        """
        params = {
            "filter": filter
        }
        return self.requester.get(params=params)

    def get_app(self, app_id):
        """ Provide information on a specific application.
        """
        return self.requester.get(app_id)

    def enable_app(self, app_id):
        """ Enable an app.
        """
        return self.requester.post(app_id)

    def disable_app(self, app_id):
        """ Disable the specified app.
        """
        return self.requester.delete(app_id)



from .requester import OCSRequester, WebDAVRequester
from .api_wrappers import OCS_API_CLASSES, WEBDAV_CLASS


class NextCloud:
    """ NextCloud API class: allows OCS and WebDav requests.
    """
    def __init__(self, endpoint, user, password, json_output=True,
                 verify=True):
        self.query_components = []

        ocs_requester = OCSRequester(
            endpoint, user, password, json_output, verify)
        webdav_requester = WebDAVRequester(
            endpoint, user, password, verify=verify)

        self.functionality_classes = [
            api_class(ocs_requester) for api_class in OCS_API_CLASSES]
        self.functionality_classes.append(
            WEBDAV_CLASS(webdav_requester, json_output=json_output))

        for functionality_class in self.functionality_classes:
            for potential_method in dir(functionality_class):
                if (potential_method.startswith('_') or not
                        callable(getattr(functionality_class,
                                         potential_method))):
                    continue
                setattr(self, potential_method,
                        getattr(functionality_class, potential_method))



from ..base import WithRequester


class Capabilities(WithRequester):
    API_URL = "/ocs/v1.php/cloud/capabilities"
    SUCCESS_CODE = 100

    def get_capabilities(self):
        """ Obtain capabilities provided by the Nextcloud server and its apps.
        """
        return self.requester.get()



from ..base import WithRequester


class Forms(WithRequester):
    API_URL = "/ocs/v2.php/apps/forms/api/v1"

    def get_forms(self):
        """ Retrieve a list of forms from the Nextcloud server.
        """
        url = "forms"
        return self.requester.get(url)



from ..base import WithRequester, ShareType


class Share(WithRequester):
    API_URL = "/ocs/v2.php/apps/files_sharing/api/v1"
    LOCAL = "shares"
    SUCCESS_CODE = 200

    def get_local_url(self, additional_url=""):
        if additional_url:
            return f"{self.LOCAL}/{additional_url}"
        return self.LOCAL

    @staticmethod
    def validate_share_parameters(path, share_type, share_with):
        """ Check if share parameters make sense.
        """
        if (path is None or not isinstance(share_type, int) or (
                share_with is None and
                share_type in [ShareType.GROUP, ShareType.USER,
                               ShareType.FEDERATED_CLOUD_SHARE,
                               ShareType.CIRCLE])):
            return False
        return True

    def get_shares(self):
        """ Get all shares from the user.
        """
        return self.requester.get(self.get_local_url())

    def get_shares_from_path(self, path, reshares=None, subfiles=None):
        """ Get all shares from a given file/folder.
        """
        url = self.get_local_url()
        params = {
            "path": path,
            # TODO: test reshares, subfiles
            "reshares": (
                None if reshares is None else str(bool(reshares)).lower()),
            "subfiles": (
                None if subfiles is None else str(bool(subfiles)).lower()),
        }
        return self.requester.get(url, params=params)

    def get_share_info(self, sid):
        """ Get information about a given share.
        """
        return self.requester.get(self.get_local_url(sid))

    def create_share(
            self, path, share_type, share_with=None, public_upload=None,
            password=None, permissions=None):
        """ Share a file/folder with a user/group or as public link.

        Mandatory fields: share_type, path and share_with for share_type USER
        (0) or GROUP (1).
        """
        if not self.validate_share_parameters(path, share_type, share_with):
            raise ValueError("Unexpected input parameters.")

        url = self.get_local_url()
        if public_upload:
            public_upload = "true"

        data = {"path": path, "shareType": share_type}
        if share_type in [ShareType.GROUP, ShareType.USER,
                          ShareType.FEDERATED_CLOUD_SHARE]:
            data["shareWith"] = share_with
        if public_upload:
            data["publicUpload"] = public_upload
        if share_type == ShareType.PUBLIC_LINK and password is not None:
            data["password"] = str(password)
        if permissions is not None:
            data["permissions"] = permissions
        return self.requester.post(url, data)

    def delete_share(self, sid):
        """ Remove the given share.
        """
        return self.requester.delete(self.get_local_url(sid))

    def update_share(self, sid, permissions=None, password=None,
                     public_upload=None, expire_date=""):
        """ Update a given share, only one value can be updated per request.
        """
        if permissions is not None:
            permissions = int(permissions)
        params = {
            "permissions": permissions,
            "password": password,
            "expireDate": expire_date
        }
        if public_upload:
            params["publicUpload"] = "true"
        if public_upload is False:
            params["publicUpload"] = "false"

        # check if only one param specified
        specified_params_count = sum([
            int(bool(each)) for each in params.values()])
        if specified_params_count > 1:
            raise ValueError("Only one parameter for update can be specified "
                             "per request")

        url = self.get_local_url(sid)
        return self.requester.put(url, data=params)



##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module defines the mzml dataset loader.
"""

# Third party import
import pymzml

# Package import
from .loader_base import LoaderBase


class MZML(LoaderBase):
    """ Define the MZML loader.
    """
    allowed_extensions = [".mzML"]

    def load(self, path):
        """ A method that load the table data.

        Parameters
        ----------
        path: str
            the path to the table to be loaded.

        Returns
        -------
        data: Ordered dict
            the loaded table.
        """

        return pymzml.run.Reader(path)



import requests
import logging
from functools import wraps
from .response import WebDAVResponse, OCSResponse
logger = logging.getLogger("rlink")


class NextCloudConnectionError(Exception):
    """ A connection error occurred.
    """


def catch_connection_error(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except requests.RequestException as e:
            raise NextCloudConnectionError(
                "Failed to establish connection to NextCloud",
                getattr(e.request, 'url', None), e)
    return wrapper


class Requester:
    def __init__(self, endpoint, user, passwd, json_output=False, verify=True):
        self.query_components = []
        self.json_output = json_output
        self.verify = verify
        self.base_url = endpoint
        self.h_get = {"OCS-APIRequest": "true"}
        self.h_post = {"OCS-APIRequest": "true",
                       "Content-Type": "application/x-www-form-urlencoded"}
        self.auth_pk = (user, passwd)
        self.API_URL = None
        self.SUCCESS_CODE = None

    def rtn(self, resp):
        if self.json_output:
            return resp.json()
        else:
            return resp.content.decode("UTF-8")

    @catch_connection_error
    def get(self, url="", params=None):
        url = self.get_full_url(url)
        logger.debug(url)
        logger.debug(params)
        res = requests.get(
            url, auth=self.auth_pk, headers=self.h_get, params=params,
            verify=self.verify)
        return self.rtn(res)

    @catch_connection_error
    def post(self, url="", data=None):
        url = self.get_full_url(url)
        logger.debug(url)
        logger.debug(data)
        res = requests.post(
            url, auth=self.auth_pk, data=data, headers=self.h_post,
            verify=self.verify)
        return self.rtn(res)

    @catch_connection_error
    def put(self, url="", data=None):
        url = self.get_full_url(url)
        logger.debug(url)
        res = requests.put(
            url, auth=self.auth_pk, data=data, headers=self.h_post,
            verify=self.verify)
        return self.rtn(res)

    @catch_connection_error
    def delete(self, url="", data=None):
        url = self.get_full_url(url)
        logger.debug(url)
        logger.debug(data)
        res = requests.delete(
            url, auth=self.auth_pk, data=data, headers=self.h_post,
            verify=self.verify)
        return self.rtn(res)

    def get_full_url(self, additional_url=""):
        """
        Build full url for request to NextCloud api

        Construct url from self.base_url, self.API_URL, additional_url
        (if given), add format=json param if self.json

        :param additional_url: str
            add to url after api_url
        :return: str
        """
        if additional_url and not str(additional_url).startswith("/"):
            additional_url = f"/{additional_url}"

        if self.json_output:
            self.query_components.append("format=json")

        ret = f"{self.base_url}{self.API_URL}{additional_url}"

        if self.json_output:
            ret += "?format=json"
        return ret


class OCSRequester(Requester):
    """ Requester for OCS API.
    """
    def rtn(self, resp):
        return OCSResponse(response=resp,
                           json_output=self.json_output,
                           success_code=self.SUCCESS_CODE)


class WebDAVRequester(Requester):
    """ Requester for WebDAV API.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def rtn(self, resp, data=None):
        return WebDAVResponse(response=resp, data=data)

    @catch_connection_error
    def propfind(self, additional_url="", headers=None, data=None):
        url = self.get_full_url(additional_url=additional_url)
        res = requests.request(
            'PROPFIND', url, auth=self.auth_pk, headers=headers, data=data,
            verify=self.verify)
        return self.rtn(res)

    @catch_connection_error
    def proppatch(self, additional_url="", data=None):
        url = self.get_full_url(additional_url=additional_url)
        res = requests.request('PROPPATCH', url, auth=self.auth_pk, data=data)
        return self.rtn(resp=res)

    @catch_connection_error
    def report(self, additional_url="", data=None):
        url = self.get_full_url(additional_url=additional_url)
        res = requests.request(
            'REPORT', url, auth=self.auth_pk, data=data, verify=self.verify)
        return self.rtn(resp=res)

    @catch_connection_error
    def download(self, url="", params=None):
        url = self.get_full_url(url)
        logger.debug(url)
        logger.debug(params)
        res = requests.get(
            url, auth=self.auth_pk, headers=self.h_get, params=params,
            verify=self.verify)
        return self.rtn(resp=res, data=res.content)

    @catch_connection_error
    def make_collection(self, additional_url=""):
        url = self.get_full_url(additional_url=additional_url)
        res = requests.request(
            "MKCOL", url=url, auth=self.auth_pk, verify=self.verify)
        return self.rtn(resp=res)

    @catch_connection_error
    def move(self, url, destination, overwrite=False):
        url = self.get_full_url(additional_url=url)
        destination_url = self.get_full_url(additional_url=destination)
        logger.debug(f"{url} -> {destination_url}")
        headers = {
            "Destination": destination_url.encode('utf-8'),
            "Overwrite": "T" if overwrite else "F"
        }
        res = requests.request(
            "MOVE", url=url, auth=self.auth_pk, headers=headers,
            verify=self.verify)
        return self.rtn(resp=res)

    @catch_connection_error
    def copy(self, url, destination, overwrite=False):
        url = self.get_full_url(additional_url=url)
        destination_url = self.get_full_url(additional_url=destination)
        logger.debug(f"{url} -> {destination_url}")
        headers = {
            "Destination": destination_url,
            "Overwrite": "T" if overwrite else "F"
        }
        res = requests.request(
            "COPY", url=url, auth=self.auth_pk, headers=headers,
            verify=self.verify)
        return self.rtn(resp=res)
